from .hdbscan import HDBSCAN, fast_hdbscan
from .branches import BranchDetector, find_branch_sub_clusters

# Force JIT compilation on import
import numpy as np
random_state = np.random.RandomState(42)
random_data = random_state.random(size=(100, 3))
HDBSCAN(allow_single_cluster=True).fit(random_data)
HDBSCAN(cluster_selection_method="leaf").fit(random_data)

__all__ = ["HDBSCAN", "fast_hdbscan", "BranchDetector", "find_branch_sub_clusters"]



import numpy as np

from sklearn.base import BaseEstimator, ClusterMixin
from sklearn.utils import check_array, check_X_y
from sklearn.utils.validation import check_is_fitted, _check_sample_weight
from sklearn.neighbors import KDTree

from warnings import warn

from .numba_kdtree import kdtree_to_numba
from .boruvka import parallel_boruvka
from .cluster_trees import (
    mst_to_linkage_tree,
    mst_to_linkage_tree_w_sample_weights,
    condense_tree,
    simplify_hierarchy,
    extract_eom_clusters,
    cluster_tree_leaves,
    cluster_epsilon_search,
    get_cluster_labelling_at_cut,
    get_cluster_label_vector,
    get_point_membership_strength_vector,
    cluster_tree_from_condensed_tree,
    extract_clusters_bcubed,
)

try:
    from hdbscan.plots import CondensedTree, SingleLinkageTree, MinimumSpanningTree

    _HAVE_HDBSCAN = True
except ImportError:
    _HAVE_HDBSCAN = False


def to_numpy_rec_array(named_tuple_tree):
    size = named_tuple_tree.parent.shape[0]
    result = np.empty(
        size,
        dtype=[
            ("parent", np.intp),
            ("child", np.intp),
            ("lambda_val", float),
            ("child_size", np.float32),
        ],
    )

    result["parent"] = named_tuple_tree.parent
    result["child"] = named_tuple_tree.child
    result["lambda_val"] = named_tuple_tree.lambda_val
    result["child_size"] = named_tuple_tree.child_size

    return result


def remap_condensed_tree(tree, internal_to_raw, outliers):
    """
    Takes an internal condensed_tree structure and adds back in a set of points
    that were initially detected as non-finite and returns that new tree.
    These points will all be split off from the maximal node at lambda zero and
    considered noise points.
    Parameters
    ----------
    tree: condensed_tree
    internal_to_raw: dict
        a mapping from internal integer index to the raw integer index
    finite_index: ndarray
        Boolean array of which entries in the raw data were finite
    """
    finite_count = len(internal_to_raw)

    outlier_count = len(outliers)
    for i, (parent, child, lambda_val, child_size) in enumerate(tree):
        if child < finite_count:
            child = internal_to_raw[child]
        else:
            child = child + outlier_count
        tree[i] = (parent + outlier_count, child, lambda_val, child_size)

    outlier_list = []
    root = tree[0][0]  # Should I check to be sure this is the minimal lambda?
    for outlier in outliers:
        outlier_list.append((root, outlier, 0, 1))

    outlier_tree = np.array(
        outlier_list,
        dtype=[
            ("parent", np.intp),
            ("child", np.intp),
            ("lambda_val", float),
            ("child_size", np.intp),
        ],
    )
    tree = np.append(outlier_tree, tree)
    return tree


def remap_single_linkage_tree(tree, internal_to_raw, outliers):
    """
    Takes an internal single_linkage_tree structure and adds back in a set of points
    that were initially detected as non-finite and returns that new tree.
    These points will all be merged into the final node at np.inf distance and
    considered noise points.
    Parameters
    ----------
    tree: single_linkage_tree
    internal_to_raw: dict
        a mapping from internal integer index to the raw integer index
    finite_index: ndarray
        Boolean array of which entries in the raw data were finite
    """
    finite_count = len(internal_to_raw)

    outlier_count = len(outliers)
    for i, (left, right, distance, size) in enumerate(tree):
        if left < finite_count:
            tree[i, 0] = internal_to_raw[left]
        else:
            tree[i, 0] = left + outlier_count
        if right < finite_count:
            tree[i, 1] = internal_to_raw[right]
        else:
            tree[i, 1] = right + outlier_count

    outlier_tree = np.zeros((len(outliers), 4))
    last_cluster_id = tree[tree.shape[0] - 1][0:2].max()
    last_cluster_size = tree[tree.shape[0] - 1][3]
    for i, outlier in enumerate(outliers):
        outlier_tree[i] = (outlier, last_cluster_id + 1, np.inf, last_cluster_size + 1)
        last_cluster_id += 1
        last_cluster_size += 1
    tree = np.vstack([tree, outlier_tree])
    return tree


def fast_hdbscan(
    data,
    data_labels=None,
    semi_supervised=False,
    ss_algorithm='bc',
    min_samples=10,
    min_cluster_size=10,
    cluster_selection_method="eom",
    max_cluster_size=np.inf,
    allow_single_cluster=False,
    cluster_selection_epsilon=0.0,
    cluster_selection_persistence=0.0,
    sample_weights=None,
    return_trees=False,
):
    data = check_array(data)

    # Detect parameter inconsistencies early.
    if semi_supervised:
        if data_labels is None:
            raise ValueError(
                "data_labels must not be None when semi_supervised is set to True!"
            )
        if ss_algorithm not in ["bc", "bc_simple"]:
            raise ValueError(f"Invalid ss_algorithm {ss_algorithm}")

    if (
        (not (np.issubdtype(type(min_samples), np.integer) or min_samples is None))
        or not np.issubdtype(type(min_cluster_size), np.integer)
        or (min_samples is not None and min_samples <= 0)
        or min_cluster_size <= 0
    ):
        raise ValueError("Min samples and min cluster size must be positive integers!")

    if (
        not np.issubdtype(type(cluster_selection_epsilon), np.floating)
        or cluster_selection_epsilon < 0.0
    ):
        raise ValueError(
            "Cluster selection epsilon must be a positive floating point number!"
        )
    if (
        not np.issubdtype(type(cluster_selection_persistence), np.floating)
        or cluster_selection_persistence < 0.0
    ):
        raise ValueError(
            "Cluster selection persistence must be a positive floating point number!"
        )

    minimum_spanning_tree, neighbors, core_distances = compute_minimum_spanning_tree(
        data,
        min_samples=min_cluster_size if min_samples is None else min_samples,
        sample_weights=sample_weights,
    )

    return (
        *clusters_from_spanning_tree(
            minimum_spanning_tree,
            data_labels=data_labels,
            semi_supervised=semi_supervised,
            ss_algorithm=ss_algorithm,
            min_cluster_size=min_cluster_size,
            cluster_selection_method=cluster_selection_method,
            max_cluster_size=max_cluster_size,
            allow_single_cluster=allow_single_cluster,
            cluster_selection_epsilon=cluster_selection_epsilon,
            cluster_selection_persistence=cluster_selection_persistence,
            sample_weights=sample_weights,
        ),
        neighbors,
        core_distances,
    )[: (None if return_trees else 2)]


def compute_minimum_spanning_tree(data, min_samples=10, sample_weights=None):
    sklearn_tree = KDTree(data)
    numba_tree = kdtree_to_numba(sklearn_tree)
    edges, neighbors, core_distances = parallel_boruvka(
        numba_tree, min_samples=min_samples, sample_weights=sample_weights
    )
    return edges, neighbors, core_distances


def clusters_from_spanning_tree(
    minimum_spanning_tree,
    data_labels=None,
    semi_supervised=False,
    ss_algorithm='bc',
    min_cluster_size=10,
    cluster_selection_method="eom",
    max_cluster_size=np.inf,
    allow_single_cluster=False,
    cluster_selection_epsilon=0.0,
    cluster_selection_persistence=0.0,
    sample_weights=None,
):
    n_points = minimum_spanning_tree.shape[0] + 1
    sorted_mst = minimum_spanning_tree[np.argsort(minimum_spanning_tree.T[2])]
    
    if sample_weights is None:
        linkage_tree = mst_to_linkage_tree(sorted_mst)
    else:
        linkage_tree = mst_to_linkage_tree_w_sample_weights(sorted_mst, sample_weights)
    
    condensed_tree = condense_tree(
        linkage_tree, min_cluster_size=min_cluster_size, sample_weights=sample_weights
    )
    if cluster_selection_persistence > 0.0:
        condensed_tree = simplify_hierarchy(
            condensed_tree, n_points, cluster_selection_persistence
        )
    
    cluster_tree = cluster_tree_from_condensed_tree(condensed_tree)
    if cluster_selection_method == "eom":
        if semi_supervised:
            # Assumes ss_algorithm is either 'bc' or 'bc_simple'
            selected_clusters = extract_clusters_bcubed(
                condensed_tree,
                cluster_tree,
                data_labels,
                allow_virtual_nodes=True if ss_algorithm == "bc" else False,
                allow_single_cluster=allow_single_cluster,
            )
        else:
            selected_clusters = extract_eom_clusters(
                condensed_tree,
                cluster_tree,
                max_cluster_size=max_cluster_size,
                allow_single_cluster=allow_single_cluster,
            )
    elif cluster_selection_method == "leaf":
        if cluster_tree.parent.shape[0] == 0:
            selected_clusters = np.empty(0, dtype=np.int64)
        else:
            selected_clusters = cluster_tree_leaves(cluster_tree, n_points)
    else:
        raise ValueError(f"Invalid cluster_selection_method {cluster_selection_method}")

    if len(selected_clusters) > 1 and cluster_selection_epsilon > 0.0:
        selected_clusters = cluster_epsilon_search(
            selected_clusters,
            cluster_tree,
            min_epsilon=cluster_selection_epsilon,
        )

    clusters = get_cluster_label_vector(
        condensed_tree,
        selected_clusters,
        cluster_selection_epsilon,
        n_samples=n_points
    )
    membership_strengths = get_point_membership_strength_vector(
        condensed_tree, selected_clusters, clusters
    )

    return clusters, membership_strengths, linkage_tree, condensed_tree, sorted_mst


class HDBSCAN(BaseEstimator, ClusterMixin):
    def __init__(
        self,
        *,
        min_cluster_size=5,
        min_samples=None,
        cluster_selection_method="eom",
        allow_single_cluster=False,
        max_cluster_size=np.inf,
        cluster_selection_epsilon=0.0,
        cluster_selection_persistence=0.0,
        semi_supervised=False,
        ss_algorithm='bc',
        **kwargs,
    ):
        self.min_cluster_size = min_cluster_size
        self.min_samples = min_samples
        self.cluster_selection_method = cluster_selection_method
        self.allow_single_cluster = allow_single_cluster
        self.max_cluster_size = max_cluster_size
        self.cluster_selection_epsilon = cluster_selection_epsilon
        self.cluster_selection_persistence = cluster_selection_persistence
        self.semi_supervised = semi_supervised
        self.ss_algorithm = ss_algorithm

    def fit(self, X, y=None, sample_weight=None, **fit_params):

        if self.semi_supervised:
            X, y = check_X_y(X, y, accept_sparse="csr", ensure_all_finite=False)
            self._raw_labels = y
            # Replace non-finite labels with -1 labels
            y[~np.isfinite(y)] = -1

            if ~np.any(y != -1):
                raise ValueError(
                    "y must contain at least one label > -1. Currently it only contains -1 and/or non-finite labels!"
                )
        else:
            X = check_array(X, accept_sparse="csr", ensure_all_finite=False)
            self._raw_data = X
        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, dtype=np.float32)

        self._all_finite = np.all(np.isfinite(X))
        if ~self._all_finite:
            # Pass only the purely finite indices into hdbscan
            # We will later assign all non-finite points to the background -1 cluster
            finite_index = np.where(np.isfinite(X).sum(axis=1) == X.shape[1])[0]
            clean_data = X[finite_index]
            clean_data_labels = y[finite_index] if self.semi_supervised else None
            sample_weight = sample_weight[finite_index] if sample_weight is not None else None

            internal_to_raw = {
                x: y for x, y in zip(range(len(finite_index)), finite_index)
            }
            outliers = list(set(range(X.shape[0])) - set(finite_index))
        else:
            clean_data = X
            clean_data_labels = y

        kwargs = self.get_params()

        (
            self.labels_,
            self.probabilities_,
            self._single_linkage_tree,
            self._condensed_tree,
            self._min_spanning_tree,
            self._neighbors,
            self._core_distances
        ) = fast_hdbscan(
            clean_data,
            clean_data_labels,
            return_trees=True,
            sample_weights=sample_weight,
            **kwargs,
        )

        self._condensed_tree = to_numpy_rec_array(self._condensed_tree)

        if not self._all_finite:
            # remap indices to align with original data in the case of non-finite entries.
            self._condensed_tree = remap_condensed_tree(
                self._condensed_tree, internal_to_raw, outliers
            )
            self._single_linkage_tree = remap_single_linkage_tree(
                self._single_linkage_tree, internal_to_raw, outliers
            )
            new_labels = np.full(X.shape[0], -1)
            new_labels[finite_index] = self.labels_
            self.labels_ = new_labels

            new_probabilities = np.zeros(X.shape[0])
            new_probabilities[finite_index] = self.probabilities_
            self.probabilities_ = new_probabilities

        return self

    def dbscan_clustering(self, epsilon):
        check_is_fitted(
            self,
            "_single_linkage_tree",
            msg="You first need to fit the HDBSCAN model before picking a DBSCAN clustering",
        )
        return get_cluster_labelling_at_cut(
            self._single_linkage_tree,
            epsilon,
            self.min_samples if self.min_samples is not None else self.min_cluster_size,
        )

    @property
    def condensed_tree_(self):
        check_is_fitted(
            self,
            "_condensed_tree",
            msg="You first need to fit the HDBSCAN model before accessing the condensed tree",
        )
        if self._condensed_tree is not None:
            return CondensedTree(
                self._condensed_tree,
                self.labels_,
            )
        else:
            raise AttributeError(
                "No condensed tree was generated; try running fit first."
            )

    @property
    def single_linkage_tree_(self):
        check_is_fitted(
            self,
            "_single_linkage_tree",
            msg="You first need to fit the HDBSCAN model before accessing the single linkage tree",
        )
        if self._single_linkage_tree is not None:
            return SingleLinkageTree(self._single_linkage_tree)
        else:
            raise AttributeError(
                "No single linkage tree was generated; try running fit first."
            )

    @property
    def minimum_spanning_tree_(self):
        check_is_fitted(
            self,
            "_min_spanning_tree",
            msg="You first need to fit the HDBSCAN model before accessing the minimum spanning tree",
        )
        if self._min_spanning_tree is not None:
            if self._raw_data is not None:
                return MinimumSpanningTree(self._min_spanning_tree, self._raw_data)
            else:
                warn(
                    "No raw data is available; this may be due to using"
                    " a precomputed metric matrix. No minimum spanning"
                    " tree will be provided without raw data."
                )
                return None
        else:
            raise AttributeError(
                "No minimum spanning tree was generated; try running fit first."
            )



import numba
import numpy as np

from collections import namedtuple

RankDisjointSet = namedtuple("DisjointSet", ["parent", "rank"])
SizeDisjointSet = namedtuple("DisjointSet", ["parent", "size"])


@numba.njit()
def ds_rank_create(n_elements):
    return RankDisjointSet(np.arange(n_elements, dtype=np.int32), np.zeros(n_elements, dtype=np.int32))


@numba.njit()
def ds_size_create(n_elements):
    return SizeDisjointSet(np.arange(n_elements, dtype=np.int32), np.ones(n_elements, dtype=np.int32))


@numba.njit()
def ds_find(disjoint_set, x):
    while disjoint_set.parent[x] != x:
        x, disjoint_set.parent[x] = disjoint_set.parent[x], disjoint_set.parent[disjoint_set.parent[x]]

    return x


@numba.njit()
def ds_union_by_rank(disjoint_set, x, y):
    x = ds_find(disjoint_set, x)
    y = ds_find(disjoint_set, y)

    if x == y:
        return

    if disjoint_set.rank[x] < disjoint_set.rank[y]:
        x, y = y, x

    disjoint_set.parent[y] = x
    if disjoint_set.rank[x] == disjoint_set.rank[y]:
        disjoint_set.rank[x] += 1


@numba.njit()
def ds_union_by_size(disjoint_set, x, y):
    x = ds_find(disjoint_set, x)
    y = ds_find(disjoint_set, y)

    if x == y:
        return

    if disjoint_set.size[x] < disjoint_set.size[y]:
        x, y = y, x

    disjoint_set.parent[y] = x
    disjoint_set.size[x] += disjoint_set.size[y]


import numba
import numpy as np

from collections import namedtuple

from .disjoint_set import ds_rank_create, ds_find, ds_union_by_rank

from numba.typed import Dict, List
from numba.types import int64, ListType

int64_list_type = ListType(int64)

LinkageMergeData = namedtuple("LinkageMergeData", ["parent", "size", "next"])


@numba.njit()
def create_linkage_merge_data(base_size):
    parent = np.full(2 * base_size - 1, -1, dtype=np.intp)
    size = np.concatenate((np.ones(base_size, dtype=np.intp), np.zeros(base_size - 1, dtype=np.intp)))
    next_parent = np.array([base_size], dtype=np.intp)

    return LinkageMergeData(parent, size, next_parent)


@numba.njit()
def create_linkage_merge_data_w_sample_weights(sample_weights):
    base_size = sample_weights.shape[0]
    parent = np.full(2 * base_size - 1, -1, dtype=np.intp)
    size = np.concatenate((sample_weights, np.zeros(base_size - 1, dtype=np.float32)))
    next_parent = np.array([base_size], dtype=np.intp)

    return LinkageMergeData(parent, size, next_parent)


@numba.njit()
def linkage_merge_find(linkage_merge, node):
    relabel = node
    while linkage_merge.parent[node] != -1 and linkage_merge.parent[node] != node:
        node = linkage_merge.parent[node]

    linkage_merge.parent[node] = node

    # label up to the root
    while linkage_merge.parent[relabel] != node:
        next_relabel = linkage_merge.parent[relabel]
        linkage_merge.parent[relabel] = node
        relabel = next_relabel

    return node


@numba.njit()
def linkage_merge_join(linkage_merge, left, right):
    linkage_merge.size[linkage_merge.next[0]] = linkage_merge.size[left] + linkage_merge.size[right]
    linkage_merge.parent[left] = linkage_merge.next[0]
    linkage_merge.parent[right] = linkage_merge.next[0]
    linkage_merge.next[0] += 1


@numba.njit()
def mst_to_linkage_tree(sorted_mst):
    result = np.empty((sorted_mst.shape[0], sorted_mst.shape[1] + 1))

    n_samples = sorted_mst.shape[0] + 1
    linkage_merge = create_linkage_merge_data(n_samples)

    for index in range(sorted_mst.shape[0]):

        left = np.intp(sorted_mst[index, 0])
        right = np.intp(sorted_mst[index, 1])
        delta = sorted_mst[index, 2]

        left_component = linkage_merge_find(linkage_merge, left)
        right_component = linkage_merge_find(linkage_merge, right)

        if left_component > right_component:
            result[index][0] = left_component
            result[index][1] = right_component
        else:
            result[index][1] = left_component
            result[index][0] = right_component

        result[index][2] = delta
        result[index][3] = linkage_merge.size[left_component] + linkage_merge.size[right_component]

        linkage_merge_join(linkage_merge, left_component, right_component)

    return result


@numba.njit()
def mst_to_linkage_tree_w_sample_weights(sorted_mst, sample_weights):
    result = np.empty((sorted_mst.shape[0], sorted_mst.shape[1] + 1))

    linkage_merge = create_linkage_merge_data_w_sample_weights(sample_weights)

    for index in range(sorted_mst.shape[0]):

        left = np.intp(sorted_mst[index, 0])
        right = np.intp(sorted_mst[index, 1])
        delta = sorted_mst[index, 2]

        left_component = linkage_merge_find(linkage_merge, left)
        right_component = linkage_merge_find(linkage_merge, right)

        if left_component > right_component:
            result[index][0] = left_component
            result[index][1] = right_component
        else:
            result[index][1] = left_component
            result[index][0] = right_component

        result[index][2] = delta
        result[index][3] = linkage_merge.size[left_component] + linkage_merge.size[right_component]

        linkage_merge_join(linkage_merge, left_component, right_component)

    return result


@numba.njit()
def bfs_from_hierarchy(hierarchy, bfs_root, num_points):
    to_process = [bfs_root]
    result = []

    while to_process:
        result.extend(to_process)
        next_to_process = []
        for n in to_process:
            if n >= num_points:
                i = n - num_points
                next_to_process.append(int(hierarchy[i, 0]))
                next_to_process.append(int(hierarchy[i, 1]))
        to_process = next_to_process

    return result


@numba.njit()
def eliminate_branch(branch_node, parent_node, lambda_value, parents, children, lambdas, sizes, idx, ignore, hierarchy,
                     num_points):
    if branch_node < num_points:
        parents[idx] = parent_node
        children[idx] = branch_node
        lambdas[idx] = lambda_value
        idx += 1
    else:
        for sub_node in bfs_from_hierarchy(hierarchy, branch_node, num_points):
            if sub_node < num_points:
                children[idx] = sub_node
                parents[idx] = parent_node
                lambdas[idx] = lambda_value
                idx += 1
            else:
                ignore[sub_node] = True

    return idx


CondensedTree = namedtuple('CondensedTree', ['parent', 'child', 'lambda_val', 'child_size'])


@numba.njit()
def empty_condensed_tree():
    parents = np.empty(shape=0, dtype=np.intp)
    others = np.empty(shape=0, dtype=np.float32)
    return CondensedTree(parents, parents, others, others)


@numba.njit(fastmath=True)
def condense_tree(hierarchy, min_cluster_size=10, max_cluster_size=np.inf, sample_weights=None):
    root = 2 * hierarchy.shape[0]
    num_points = hierarchy.shape[0] + 1
    next_label = num_points + 1

    node_list = bfs_from_hierarchy(hierarchy, root, num_points)

    relabel = np.zeros(root + 1, dtype=np.int64)
    relabel[root] = num_points

    parents = np.ones(root, dtype=np.int64)
    children = np.empty(root, dtype=np.int64)
    lambdas = np.empty(root, dtype=np.float32)
    sizes = np.ones(root, dtype=np.float32)

    ignore = np.zeros(root + 1, dtype=np.bool_) # 'bool' is no longer an attribute of 'numpy'

    if sample_weights is None:
        sample_weights = np.ones(num_points, dtype=np.float32)

    idx = 0

    for node in node_list:
        if ignore[node] or node < num_points:
            continue

        parent_node = relabel[node]
        l, r, d, _ = hierarchy[node - num_points]
        left = np.int64(l)
        right = np.int64(r)
        if d > 0.0:
            lambda_value = 1.0 / d
        else:
            lambda_value = np.inf

        left_count = np.float32(hierarchy[left - num_points, 3]) if left >= num_points else sample_weights[left]
        right_count = np.float32(hierarchy[right - num_points, 3]) if right >= num_points else sample_weights[right]

        # The logic here is in a strange order, but it has non-trivial performance gains ...
        # The most common case by far is a singleton on the left; and cluster on the right take care of this separately
        if left < num_points and right_count >= min_cluster_size:
            relabel[right] = parent_node
            parents[idx] = parent_node
            children[idx] = left
            lambdas[idx] = lambda_value
            idx += 1
        # Next most common is a small left cluster and a large right cluster: relabel the right node; eliminate the left branch
        elif left_count < min_cluster_size and right_count >= min_cluster_size:
            relabel[right] = parent_node
            idx = eliminate_branch(left, parent_node, lambda_value, parents, children, lambdas, sizes, idx, ignore,
                                   hierarchy, num_points)
        # Then we have a large left cluster and a small right cluster: relabel the left node; eliminate the right branch
        elif left_count >= min_cluster_size and right_count < min_cluster_size:
            relabel[left] = parent_node
            idx = eliminate_branch(right, parent_node, lambda_value, parents, children, lambdas, sizes, idx, ignore,
                                   hierarchy, num_points)
        # If both clusters are small then eliminate all branches
        elif left_count < min_cluster_size and right_count < min_cluster_size:
            idx = eliminate_branch(left, parent_node, lambda_value, parents, children, lambdas, sizes, idx, ignore,
                                   hierarchy, num_points)
            idx = eliminate_branch(right, parent_node, lambda_value, parents, children, lambdas, sizes, idx, ignore,
                                   hierarchy, num_points)
        # If both clusters are too large then relabel both
        elif left_count > max_cluster_size and right_count > max_cluster_size:
            relabel[left] = parent_node
            relabel[right] = parent_node
        else:
            relabel[left] = next_label

            parents[idx] = parent_node
            children[idx] = next_label
            lambdas[idx] = lambda_value
            sizes[idx] = left_count
            next_label += 1
            idx += 1

            relabel[right] = next_label

            parents[idx] = parent_node
            children[idx] = next_label
            lambdas[idx] = lambda_value
            sizes[idx] = right_count
            next_label += 1
            idx += 1

    return CondensedTree(parents[:idx], children[:idx], lambdas[:idx], sizes[:idx])


@numba.njit()
def extract_leaves(condensed_tree, allow_single_cluster=True):
    n_nodes = condensed_tree.parent.max() + 1
    n_points = condensed_tree.parent.min()
    leaf_indicator = np.ones(n_nodes, dtype=np.bool_)
    leaf_indicator[:n_points] = False

    for parent, child_size in zip(condensed_tree.parent, condensed_tree.child_size):
        if child_size > 1:
            leaf_indicator[parent] = False

    return np.nonzero(leaf_indicator)[0]


@numba.njit()
def cluster_tree_leaves(cluster_tree, n_points):
    n_nodes = cluster_tree.child.max() + 1
    leaf_indicator = np.ones(n_nodes - n_points, dtype=np.bool_)
    leaf_indicator[cluster_tree.parent - n_points] = False
    return np.nonzero(leaf_indicator)[0] + n_points


# The *_bcubed functions below implement the (semi-supervised) HDBSCAN*(BC) algorithm presented
# in Castro Gertrudes, J., Zimek, A., Sander, J. et al. A unified view of density-based methods 
# for semi-supervised clustering and classification. Data Min Knowl Disc 33, 1894–1952 (2019).

@numba.njit()
def cluster_tree_from_condensed_tree_bcubed(condensed_tree, cluster_tree, label_indices):
    # This functions returns a cluster_tree with virtual nodes (if applicable).

    label_indices_list = list(label_indices.keys())
    cluster_tree_parents = list(cluster_tree.parent)

    # A labeled node that has no children and who's parent is not a leaf cluster, then it must be 
    # a noisy node (virtual node). 

    mask1 = condensed_tree.child_size > 1
    mask2 = condensed_tree.child_size == 1
    mask3 = np.array([child in label_indices_list for child in condensed_tree.child])
    mask4 = np.array([parent in cluster_tree_parents for parent in condensed_tree.parent]) # check that it's not a leaf cluster

    mask = (mask1 | (mask2 & mask3 & mask4)) 

    return CondensedTree(condensed_tree.parent[mask], condensed_tree.child[mask], condensed_tree.lambda_val[mask],
                         condensed_tree.child_size[mask])


@numba.njit()
def get_condensed_tree_clusters_bcubed(condensed_tree, label_indices, cluster_tree=None, cluster_tree_bcubed=None, allow_virtual_nodes=False):

    cluster_elements = Dict.empty(
        key_type=int64,
        value_type=int64_list_type,
        )
    
    virtual_nodes = [0 for x in range(0)] 
    labeled_points = set(label_indices.keys())

    parents_set = set(list(condensed_tree.parent))
    for i in range(len(condensed_tree.child) - 1, -1, -1): # Traverse tree bottom up
        parent = condensed_tree.parent[i]
        child = condensed_tree.child[i]
        if child in parents_set:
            if parent in cluster_elements:
                cluster_elements[parent].extend(cluster_elements[child])
            else:
                cluster_labeled_points = list(set(cluster_elements[child]) & labeled_points)
                cluster_elements[parent] = List(cluster_labeled_points)
        elif parent in cluster_elements:
            if child in labeled_points:
                cluster_elements[parent].append(child)
        else:
            cluster_elements[parent] = List.empty_list(int64)
            if child in labeled_points:
                cluster_elements[parent].append(child)

    if allow_virtual_nodes and (cluster_tree is not None) and (cluster_tree_bcubed is not None):
        for node in list(set(cluster_tree_bcubed.child).difference(set(cluster_tree.child))):
            virtual_nodes.append(node)
            cluster_elements[node] = List.empty_list(int64)
            cluster_elements[node].append(node)

    return cluster_elements, np.array(virtual_nodes)


@numba.njit()
def eom_recursion_bcubed(node, cluster_tree, stability_node_scores, bcubed_node_scores, selected_clusters, unselected_nodes):

    current_score_stability_bcubed = np.array([stability_node_scores[node], bcubed_node_scores[node]], dtype=np.float32)

    children = cluster_tree.child[cluster_tree.parent == node]
    child_score_total_stability_bcubed = np.array([0.0, 0.0], dtype=np.float32)

    for child_node in children:
        child_score_total_stability_bcubed += eom_recursion_bcubed(child_node, cluster_tree, stability_node_scores, bcubed_node_scores, selected_clusters, unselected_nodes)

    if child_score_total_stability_bcubed[1] > current_score_stability_bcubed[1]:
        return child_score_total_stability_bcubed

    elif child_score_total_stability_bcubed[1] < current_score_stability_bcubed[1]:
        selected_clusters[node] = True
        unselect_below_node_bcubed(node, cluster_tree, selected_clusters, unselected_nodes)
        return current_score_stability_bcubed   

    # Stability scores used to resolve ties.
    elif child_score_total_stability_bcubed[1] == current_score_stability_bcubed[1]:
        
        if child_score_total_stability_bcubed[0] > current_score_stability_bcubed[0]:
            return child_score_total_stability_bcubed
        else:
            selected_clusters[node] = True
            unselect_below_node_bcubed(node, cluster_tree, selected_clusters, unselected_nodes)
            return current_score_stability_bcubed


@numba.njit()
def score_condensed_tree_nodes_bcubed(cluster_elements, label_indices): 

    label_values = label_indices.values()
    label_counts = {0: 0 for i in range(0)}

    for label in label_values:
        if label in label_counts:
            label_counts[label] +=1
        else:
            label_counts[label] = 1

    label_counts_values = list(label_counts.values())
    total_num_of_labeled_points = sum(label_counts_values)
    bcubed = {0: 0.0 for i in range(0)}

    for cluster, cluster_labeled_points in cluster_elements.items():

        cluster_labeled_points_dict = {0: 0 for i in range(0)}
        bcubed[cluster] = 0.0

        if len(cluster_labeled_points) > 0:
            
            for p in cluster_labeled_points:
                p_label = label_indices[p]
                if p_label in cluster_labeled_points_dict:
                    cluster_labeled_points_dict[p_label] += 1
                else:
                    cluster_labeled_points_dict[p_label] = 1
    
            for label, num_points in cluster_labeled_points_dict.items():

                total_num_of_class_label = label_counts[label]
                num_labeled_in_node = len(cluster_labeled_points)

                precision_point = (num_points/num_labeled_in_node)/total_num_of_labeled_points
                recall_point = (num_points/total_num_of_class_label)/total_num_of_labeled_points

                # Bcubed F-measure 
                bcubed[cluster] += num_points*(2.0/(1.0/precision_point + 1.0/recall_point))
    return bcubed

@numba.njit()
def unselect_below_node_bcubed(node, cluster_tree, selected_clusters, unselected_nodes):
    
    for child in cluster_tree.child[cluster_tree.parent == node]:
        if not unselected_nodes[child]:
            unselect_below_node_bcubed(child, cluster_tree, selected_clusters, unselected_nodes)
            selected_clusters[child] = False
            unselected_nodes[child] = True

@numba.njit()
def extract_clusters_bcubed(condensed_tree, cluster_tree, data_labels, allow_virtual_nodes=False, allow_single_cluster=False):
    label_indices = Dict()
    for index in np.flatnonzero(data_labels > -1):
        label_indices[index] = data_labels[index]

    if allow_virtual_nodes:

        cluster_tree_bcubed = cluster_tree_from_condensed_tree_bcubed(condensed_tree, cluster_tree, label_indices)
        cluster_elements, virtual_nodes = get_condensed_tree_clusters_bcubed(condensed_tree, label_indices, cluster_tree, cluster_tree_bcubed, allow_virtual_nodes)
        stability_node_scores = score_condensed_tree_nodes(condensed_tree)
        for node in virtual_nodes:
            stability_node_scores[node] = np.float32(0.0)
        bcubed_node_scores = score_condensed_tree_nodes_bcubed(cluster_elements, label_indices)
              
    else:

        cluster_tree_bcubed = cluster_tree
        cluster_elements, virtual_nodes = get_condensed_tree_clusters_bcubed(condensed_tree, label_indices)
        stability_node_scores = score_condensed_tree_nodes(condensed_tree) 
        bcubed_node_scores = score_condensed_tree_nodes_bcubed(cluster_elements, label_indices)

    selected_clusters = {node: False for node in bcubed_node_scores}

    if len(cluster_tree_bcubed.parent) == 0:
        return np.zeros(0, dtype=np.int64)

    cluster_tree_root = cluster_tree_bcubed.parent.min()
    unselected_nodes = {node: False for node in bcubed_node_scores}

    if allow_single_cluster:
        eom_recursion_bcubed(cluster_tree_root, cluster_tree_bcubed, stability_node_scores, bcubed_node_scores, selected_clusters, unselected_nodes)
    elif len(bcubed_node_scores) > 1:
        root_children = cluster_tree_bcubed.child[cluster_tree_bcubed.parent == cluster_tree_root]
        for child_node in root_children:
            eom_recursion_bcubed(child_node, cluster_tree_bcubed, stability_node_scores, bcubed_node_scores, selected_clusters, unselected_nodes)

    return np.asarray([node for node, selected in selected_clusters.items() if (selected and (node not in virtual_nodes))])


@numba.njit()
def score_condensed_tree_nodes(condensed_tree):
    root = condensed_tree.parent[0]
    result = {root: np.float32(0.0)}

    for i in range(condensed_tree.parent.shape[0]):
        if condensed_tree.child_size[i] > 1:
            child = condensed_tree.child[i]
            result[child] = -condensed_tree.lambda_val[i] * condensed_tree.child_size[i]

        parent = condensed_tree.parent[i]
        result[parent] += condensed_tree.lambda_val[i] * condensed_tree.child_size[i]

    return result


@numba.njit()
def cluster_tree_from_condensed_tree(condensed_tree):
    return mask_condensed_tree(condensed_tree, condensed_tree.child_size > 1)


@numba.njit()
def mask_condensed_tree(condensed_tree, mask):
    return CondensedTree(
        condensed_tree.parent[mask], 
        condensed_tree.child[mask], 
        condensed_tree.lambda_val[mask],
        condensed_tree.child_size[mask]
    )


@numba.njit()
def unselect_below_node(node, cluster_tree, selected_clusters):
    for child in cluster_tree.child[cluster_tree.parent == node]:
        unselect_below_node(child, cluster_tree, selected_clusters)
        selected_clusters[child] = False


@numba.njit(fastmath=True)
def eom_recursion(node, cluster_tree, node_scores, node_sizes, selected_clusters, max_cluster_size):
    current_score = max(node_scores[node], 0.0) # floating point errors can make score negative!
    current_size = node_sizes[node]

    children = cluster_tree.child[cluster_tree.parent == node]
    child_score_total = 0.0

    for child_node in children:
        child_score_total += eom_recursion(child_node, cluster_tree, node_scores, node_sizes, selected_clusters, max_cluster_size)

    if child_score_total > current_score or current_size > max_cluster_size:
        return child_score_total
    else:
        selected_clusters[node] = True
        unselect_below_node(node, cluster_tree, selected_clusters)
        return current_score


@numba.njit()
def extract_eom_clusters(condensed_tree, cluster_tree, max_cluster_size=np.inf, allow_single_cluster=False):
    node_scores = score_condensed_tree_nodes(condensed_tree)
    if len(cluster_tree.parent) > 0:
        node_sizes = {node: size for node, size in zip(cluster_tree.child, cluster_tree.child_size.astype(np.float32))}
        node_sizes[cluster_tree.parent.min()] = np.float32(cluster_tree.parent.min() - 1)
    else:
        node_sizes = {-1: np.float32(0.0)}
    selected_clusters = {node: False for node in node_scores}

    if len(cluster_tree.parent) == 0:
        return np.zeros(0, dtype=np.int64)

    cluster_tree_root = cluster_tree.parent.min()

    if allow_single_cluster:
        eom_recursion(cluster_tree_root, cluster_tree, node_scores, node_sizes, selected_clusters, max_cluster_size)
    elif len(node_scores) > 1:
        root_children = cluster_tree.child[cluster_tree.parent == cluster_tree_root]
        for child_node in root_children:
            eom_recursion(child_node, cluster_tree, node_scores, node_sizes, selected_clusters, max_cluster_size)

    return np.asarray([node for node, selected in selected_clusters.items() if selected])


@numba.njit()
def simplify_hierarchy(condensed_tree, n_points, persistence_threshold):
    keep_mask = np.ones(condensed_tree.parent.shape[0], dtype=np.bool_)
    cluster_tree = cluster_tree_from_condensed_tree(condensed_tree)

    processed = {np.int64(0)}
    processed.clear()
    while cluster_tree.parent.shape[0] > 0:
        leaves = set(cluster_tree_leaves(cluster_tree, n_points))
        births = max_lambdas(condensed_tree, leaves)
        deaths = min_lambdas(cluster_tree, leaves)

        cluster_mask = np.ones(cluster_tree.parent.shape[0], dtype=np.bool_)
        for leaf in sorted(leaves, reverse=True):
            if leaf in processed or (births[leaf] - deaths[leaf]) >= persistence_threshold:
                continue
            
            # Find rows for leaf and sibling
            leaf_idx = np.searchsorted(cluster_tree.child, leaf)
            parent = cluster_tree.parent[leaf_idx]
            if leaf_idx > 0 and cluster_tree.parent[leaf_idx - 1] == parent:
                sibling_idx = leaf_idx - 1 
            else:
                sibling_idx = leaf_idx + 1
            sibling = cluster_tree.child[sibling_idx]
                        
            # Update parent values to the new parent
            for idx, row in enumerate(cluster_tree.parent):
                if row in [leaf, sibling]:
                    cluster_tree.parent[idx] = parent
            for idx, row in enumerate(condensed_tree.parent):
                if row in [leaf, sibling]:
                    condensed_tree.parent[idx] = parent
                    condensed_tree.lambda_val[idx] = deaths[leaf]
            
            # Mark visited rows
            processed.add(leaf)
            processed.add(sibling)
            cluster_mask[leaf_idx] = False
            cluster_mask[sibling_idx] = False
            for idx, child in enumerate(condensed_tree.child):
                if child in [leaf, sibling]:
                    keep_mask[idx] = False

        if np.all(cluster_mask):
            break
        cluster_tree = mask_condensed_tree(cluster_tree, cluster_mask)

    condensed_tree = mask_condensed_tree(condensed_tree, keep_mask)
    return remap_cluster_ids(condensed_tree, n_points)


@numba.njit()
def remap_cluster_ids(condensed_tree, n_points):
    n_nodes = condensed_tree.parent.max() + 1
    remaining_parents = np.unique(condensed_tree.parent)
    id_map = np.empty(n_nodes - n_points, dtype=np.int64)
    id_map[remaining_parents - n_points] = np.arange(
        n_points, n_points + remaining_parents.shape[0]
    )
    for column in [condensed_tree.parent, condensed_tree.child]:
        for idx, node in enumerate(column):
            if node >= n_points:
                column[idx] = id_map[node - n_points]
    return condensed_tree


@numba.njit()
def cluster_epsilon_search(clusters, cluster_tree, min_epsilon=0.0):
    selected = list()
    # only way to create a typed empty set
    processed = {np.int64(0)}
    processed.clear()

    # cluster_tree is sorted with increasing children
    # prepare to use binary search on parent in segment_in_branches
    parent_order = np.argsort(cluster_tree.parent)
    parents = cluster_tree.parent[parent_order]
    children = cluster_tree.child[parent_order]

    root = cluster_tree.parent.min()
    for cluster in clusters:
        idx = np.searchsorted(cluster_tree.child, cluster)
        death_eps = 1 / cluster_tree.lambda_val[idx]
        if death_eps < min_epsilon:
            if cluster not in processed:
                parent = traverse_upwards(cluster_tree, min_epsilon, root, cluster)
                selected.append(parent)
                processed |= segments_in_branch(parents, children, parent)
        else:
            selected.append(cluster)
    return np.asarray(selected)


@numba.njit()
def traverse_upwards(cluster_tree, min_epsilon, root, segment):
    parent = cluster_tree.parent[cluster_tree.child == segment][0]
    if parent == root:
        return root
    parent_death_eps = 1 / cluster_tree.lambda_val[cluster_tree.child == parent][0]
    if parent_death_eps >= min_epsilon:
        return parent
    else:
        return traverse_upwards(cluster_tree, min_epsilon, root, parent)


@numba.njit()
def segments_in_branch(parents, children, segment):
    # only way to create a typed empty set
    child_set = {np.int64(0)}
    result = {np.int64(0)}
    result.clear()
    to_process = {segment}

    while len(to_process) > 0:
        result |= to_process
   
        child_set.clear()
        for segment in to_process:
            idx = np.searchsorted(parents, segment)
            if idx >= len(parents):
                continue
            child_set.add(children[idx])
            child_set.add(children[idx + 1])
        
        to_process.clear()
        to_process |= child_set

    return result


@numba.njit(parallel=True)
def get_cluster_labelling_at_cut(linkage_tree, cut, min_cluster_size):

    root = 2 * linkage_tree.shape[0]
    num_points = linkage_tree.shape[0] + 1
    result = np.empty(num_points, dtype=np.intp)
    disjoint_set = ds_rank_create(root + 1)

    cluster = num_points
    for i in range(linkage_tree.shape[0]):
        if linkage_tree[i, 2] < cut:
            ds_union_by_rank(disjoint_set, np.intp(linkage_tree[i, 0]), cluster)
            ds_union_by_rank(disjoint_set, np.intp(linkage_tree[i, 1]), cluster)
        cluster += 1

    cluster_size = np.zeros(cluster, dtype=np.intp)
    for n in range(num_points):
        cluster = ds_find(disjoint_set, n)
        cluster_size[cluster] += 1
        result[n] = cluster

    cluster_label_map = {-1: -1}
    cluster_label = 0
    unique_labels = np.unique(result)

    for cluster in unique_labels:
        if cluster_size[cluster] < min_cluster_size:
            cluster_label_map[cluster] = -1
        else:
            cluster_label_map[cluster] = cluster_label
            cluster_label += 1

    for n in numba.prange(num_points):
        result[n] = cluster_label_map[result[n]]

    return result

@numba.njit()
def get_cluster_label_vector(
        tree,
        clusters,
        cluster_selection_epsilon,
        n_samples,
):
    if len(tree.parent) == 0 or len(clusters) == 0:
        return np.full(n_samples, -1, dtype=np.intp)
    root_cluster = tree.parent.min()
    result = np.full(n_samples, -1, dtype=np.intp)
    cluster_label_map = {c: n for n, c in enumerate(np.sort(clusters))}

    disjoint_set = ds_rank_create(max(tree.parent.max() + 1, tree.child.max() + 1))
    clusters = set(clusters)

    for n in range(tree.parent.shape[0]):
        child = tree.child[n]
        parent = tree.parent[n]
        if child not in clusters:
            ds_union_by_rank(disjoint_set, parent, child)

    for n in range(root_cluster):
        cluster = ds_find(disjoint_set, n)
        if cluster < root_cluster:
            result[n] = -1
        elif cluster == root_cluster:
            if len(clusters) == 1:
                max_lambda = tree.lambda_val[tree.parent == cluster].max()
                cur_lambda = tree.lambda_val[tree.child == n]
                if cluster_selection_epsilon > 0.0:
                    if cur_lambda >= 1 / cluster_selection_epsilon:
                        result[n] = cluster_label_map[cluster]
                    else:
                        result[n] = -1
                elif cur_lambda >= max_lambda:
                    result[n] = cluster_label_map[cluster]
                else:
                    result[n] = -1
            else:
                result[n] = -1
        else:
            result[n] = cluster_label_map[cluster]

    return result


@numba.njit()
def max_lambdas(tree, clusters):
    result = {c: 0.0 for c in clusters}

    for n in range(tree.parent.shape[0]):
        cluster = tree.parent[n]
        if cluster in clusters and tree.child_size[n] == 1:
            result[cluster] = max(result[cluster], tree.lambda_val[n])

    return result


@numba.njit()
def min_lambdas(cluster_tree, clusters):
    return {
        c: cluster_tree.lambda_val[np.searchsorted(cluster_tree.child, c)] 
        for c in clusters
    }


@numba.njit()
def get_point_membership_strength_vector(tree, clusters, labels):
    result = np.zeros(labels.shape[0], dtype=np.float32)
    deaths = max_lambdas(tree, set(clusters))
    root_cluster = tree.parent.min()
    cluster_index_map = {n: c for n, c in enumerate(np.sort(clusters))}

    for n in range(tree.child.shape[0]):
        point = tree.child[n]
        if point >= root_cluster or labels[point] < 0:
            continue

        cluster = cluster_index_map[labels[point]]
        max_lambda = deaths[cluster]
        if max_lambda == 0.0 or not np.isfinite(tree.lambda_val[n]):
            result[point] = 1.0
        else:
            lambda_val = min(tree.lambda_val[n], max_lambda)
            result[point] = lambda_val / max_lambda

    return result



"""
Converts data point lens values into edge distances and looks for clusters
induced by those distances within the clusters found by HDBSCAN.
"""

import numba
import numpy as np
from sklearn.utils.validation import check_is_fitted, _check_sample_weight
from sklearn.base import BaseEstimator, ClusterMixin

from .hdbscan import to_numpy_rec_array
from .core_graph import core_graph_clusters, core_graph_to_edge_list


def compute_sub_clusters_in_cluster(
    cluster,
    data,
    labels,
    probabilities,
    neighbors,
    core_distances,
    min_spanning_tree,
    parent_labels,
    child_labels,
    lens_callback,
    sample_weights=None,
    **kwargs,
):
    # Convert to within cluster indices (-1 indicates invalid neighbor)
    points = np.nonzero(labels == cluster)[0]

    in_cluster_ids = np.full(data.shape[0], -1, dtype=np.int32)
    in_cluster_ids[points] = np.arange(len(points))
    neighbors = in_cluster_ids[neighbors[points, :]]
    core_distances = core_distances[points]
    min_spanning_tree = min_spanning_tree[
        (parent_labels == cluster) & (child_labels == cluster)
    ]
    min_spanning_tree[:, :2] = in_cluster_ids[min_spanning_tree[:, :2].astype(np.int64)]

    # Compute lens_values
    lens_values = lens_callback(
        data, probabilities, neighbors, core_distances, min_spanning_tree, points
    )

    # Compute branches from core graph
    return (
        *core_graph_clusters(
            lens_values,
            neighbors,
            core_distances,
            min_spanning_tree,
            sample_weights=(
                sample_weights[points] if sample_weights is not None else None
            ),
            **kwargs,
        ),
        lens_values,
        points,
    )


def compute_sub_clusters_per_cluster(
    data,
    labels,
    probabilities,
    neighbors,
    core_distances,
    min_spanning_tree,
    lens_callback,
    num_clusters,
    **kwargs,
):
    # Loop could be parallel over clusters, but njit-compiling all called
    # functions slows down imports with a factor > 2 for small gains. Instead,
    # parts of each loop are parallel over points in the clusters.
    parent_labels = labels[min_spanning_tree[:, 0].astype(np.int64)]
    child_labels = labels[min_spanning_tree[:, 1].astype(np.int64)]
    return [
        compute_sub_clusters_in_cluster(
            cluster,
            data,
            labels,
            probabilities,
            neighbors,
            core_distances,
            min_spanning_tree,
            parent_labels,
            child_labels,
            lens_callback,
            **kwargs,
        )
        for cluster in range(num_clusters)
    ]


def update_labels(
    cluster_probabilities,
    sub_labels_list,
    sub_probabilities_list,
    lens_values_list,
    points_list,
    data_size,
):
    labels = np.full(data_size, -1, dtype=np.int64)
    probabilities = cluster_probabilities.copy()
    sub_labels = np.zeros(data_size, dtype=np.int64)
    sub_probabilities = np.zeros(data_size, dtype=np.float32)
    lens_values = np.zeros(data_size, dtype=np.float32)

    running_id = 0
    for points, _labels, _probs, _lens in zip(
        points_list,
        sub_labels_list,
        sub_probabilities_list,
        lens_values_list,
    ):
        unique_labels = np.unique(_labels)
        labels[points] = _labels + int(unique_labels[0] == -1) + running_id
        sub_labels[points] = _labels
        sub_probabilities[points] = _probs
        probabilities[points] += _probs
        probabilities[points] /= 2
        lens_values[points] = _lens
        running_id += len(unique_labels)

    return labels, probabilities, sub_labels, sub_probabilities, lens_values


@numba.njit()
def propagate_labels_per_cluster(graph, sub_labels):
    # create undirected core graph
    undirected = [
        {np.int64(0): np.float64(0.0) for _ in range(0)} for _ in range(len(sub_labels))
    ]
    for idx, (start, end) in enumerate(zip(graph.indptr, graph.indptr[1:])):
        for i in range(start, end):
            neighbor = graph.indices[i]
            if sub_labels[neighbor] == -1:
                undirected[neighbor][idx] = 1 / graph.weights[i]
            if sub_labels[idx] == -1:
                undirected[idx][neighbor] = 1 / graph.weights[i]

    # repeat density-weighted majority votes on noise points until all are assigned
    while True:
        noise_idx = np.nonzero(sub_labels == -1)[0]
        if noise_idx.shape[0] == 0:
            break
        for idx in noise_idx:
            candidates = {np.int64(0): np.float64(0.0) for _ in range(0)}
            for neighbor_idx, weight in undirected[idx].items():
                label = sub_labels[neighbor_idx]
                if label == -1:
                    continue
                candidates[label] = candidates.get(label, 0.0) + weight

            if len(candidates) == 0:
                continue
            max_weight = -np.inf
            max_candidate = -1
            for candidate, weight in candidates.items():
                if weight > max_weight:
                    max_weight = weight
                    max_candidate = candidate
            sub_labels[idx] = max_candidate
    return sub_labels


def propagate_sub_cluster_labels(labels, sub_labels, graph_list, points_list):
    running_id = 0
    for points, core_graph in zip(
        points_list,
        graph_list,
    ):
        unique_sub_labels = np.unique(sub_labels[points])
        has_noise = unique_sub_labels[0] == -1 and len(unique_sub_labels) > 1
        if has_noise:
            sub_labels[points] = propagate_labels_per_cluster(
                core_graph, sub_labels[points]
            )
        labels[points] = sub_labels[points] + running_id
        running_id += len(unique_sub_labels) - int(has_noise)

    return labels, sub_labels


def remap_results(
    labels,
    probabilities,
    cluster_labels,
    cluster_probabilities,
    sub_labels,
    sub_probabilities,
    lens_values,
    points,
    finite_index,
    num_points,
):
    new_labels = np.full(num_points, -1, dtype=labels.dtype)
    new_labels[finite_index] = labels
    labels = new_labels

    new_probabilities = np.full(num_points, 0.0, dtype=probabilities.dtype)
    new_probabilities[finite_index] = probabilities
    probabilities = new_probabilities

    new_cluster_labels = np.full(num_points, -1, dtype=cluster_labels.dtype)
    new_cluster_labels[finite_index] = cluster_labels
    cluster_labels = new_cluster_labels

    new_cluster_probabilities = np.full(
        num_points, 0.0, dtype=cluster_probabilities.dtype
    )
    new_cluster_probabilities[finite_index] = cluster_probabilities
    cluster_probabilities = new_cluster_probabilities

    new_sub_labels = np.full(num_points, 0, dtype=sub_labels.dtype)
    new_sub_labels[finite_index] = sub_labels
    sub_labels = new_sub_labels

    new_sub_probabilities = np.full(num_points, 1.0, dtype=sub_probabilities.dtype)
    new_sub_probabilities[finite_index] = sub_probabilities
    sub_probabilities = new_sub_probabilities

    new_lens_values = np.full(num_points, 0.0, dtype=lens_values.dtype)
    new_lens_values[finite_index] = lens_values
    lens_values = new_lens_values

    for pts in points:
        pts[:] = finite_index[pts]

    return (
        labels,
        probabilities,
        cluster_labels,
        cluster_probabilities,
        sub_labels,
        sub_probabilities,
        lens_values,
        points,
    )


def find_sub_clusters(
    clusterer,
    cluster_labels=None,
    cluster_probabilities=None,
    sample_weights=None,
    lens_callback=None,
    *,
    min_cluster_size=None,
    max_cluster_size=None,
    allow_single_cluster=None,
    cluster_selection_method=None,
    cluster_selection_epsilon=0.0,
    cluster_selection_persistence=0.0,
    propagate_labels=False,
):
    check_is_fitted(
        clusterer,
        "_min_spanning_tree",
        msg="You first need to fit the HDBSCAN model before detecting sub clusters.",
    )

    # Validate input parameters
    if cluster_labels is None:
        cluster_labels = clusterer.labels_
    elif cluster_probabilities is None:
        cluster_probabilities = np.ones(cluster_labels.shape[0], dtype=np.float32)
    if cluster_probabilities is None:
        cluster_probabilities = clusterer.probabilities_
    if sample_weights is not None:
        sample_weights = _check_sample_weight(
            sample_weights, clusterer._raw_data, dtype=np.float32
        )
    if min_cluster_size is None:
        min_cluster_size = clusterer.min_cluster_size
    if max_cluster_size is None:
        max_cluster_size = clusterer.max_cluster_size
    if allow_single_cluster is None:
        allow_single_cluster = clusterer.allow_single_cluster
    if cluster_selection_method is None:
        cluster_selection_method = clusterer.cluster_selection_method

    if not (
        np.issubdtype(type(min_cluster_size), np.integer) and min_cluster_size >= 2
    ):
        raise ValueError(
            f"min_cluster_size must be an integer greater or equal "
            f"to 2,  {min_cluster_size} given."
        )
    if max_cluster_size <= 0:
        raise ValueError(
            f"max_cluster_size must be greater 0, {max_cluster_size} given."
        )
    if not (
        np.issubdtype(type(cluster_selection_persistence), np.floating)
        and cluster_selection_persistence >= 0.0
    ):
        raise ValueError(
            f"cluster_selection_persistence must be a float greater or equal to "
            f"0.0, {cluster_selection_persistence} given."
        )
    if not (
        np.issubdtype(type(cluster_selection_epsilon), np.floating)
        and cluster_selection_epsilon >= 0.0
    ):
        raise ValueError(
            f"cluster_selection_epsilon must be a float greater or equal to "
            f"0.0, {cluster_selection_epsilon} given."
        )
    if cluster_selection_method not in ("eom", "leaf"):
        raise ValueError(
            f"Invalid cluster_selection_method: {cluster_selection_method}\n"
            f'Should be one of: "eom", "leaf"\n'
        )
    if np.all(cluster_labels == -1):
        raise ValueError("Input contains only noise points.")

    # Recover finite data points
    data = clusterer._raw_data
    num_points = data.shape[0]
    last_outlier = np.searchsorted(
        clusterer._condensed_tree["lambda_val"], 0.0, side="right"
    )
    if last_outlier > 0:
        finite_index = np.setdiff1d(
            np.arange(data.shape[0]), clusterer._condensed_tree["child"][:last_outlier]
        )
        data = data[finite_index]
        cluster_labels = cluster_labels[finite_index]
        cluster_probabilities = cluster_probabilities[finite_index]
        sample_weights = (
            sample_weights[finite_index] if sample_weights is not None else None
        )

    # Convert lens value array to callback
    if isinstance(lens_callback, np.ndarray):
        if len(lens_callback) != num_points:
            raise ValueError(
                "when providing values as lens_callback, they must have"
                f"the same length as the data, {len(lens_callback)} != {data.shape[0]}"
            )
        if last_outlier > 0:
            lens_values = lens_callback[finite_index]
        else:
            lens_values = lens_callback

        lens_callback = lambda a, b, c, d, e, pts: lens_values[pts]

    # Compute per-cluster sub clusters
    num_clusters = np.max(cluster_labels) + 1
    (
        sub_labels,
        sub_probabilities,
        linkage_trees,
        condensed_trees,
        spanning_trees,
        core_graphs,
        lens_values,
        points,
    ) = zip(
        *compute_sub_clusters_per_cluster(
            data,
            cluster_labels,
            cluster_probabilities,
            clusterer._neighbors,
            clusterer._core_distances,
            clusterer._min_spanning_tree,
            lens_callback,
            num_clusters,
            min_cluster_size=min_cluster_size,
            max_cluster_size=max_cluster_size,
            allow_single_cluster=allow_single_cluster,
            cluster_selection_method=cluster_selection_method,
            cluster_selection_epsilon=cluster_selection_epsilon,
            cluster_selection_persistence=cluster_selection_persistence,
            sample_weights=sample_weights,
        )
    )

    # Handle override labels failure cases
    condensed_trees = [
        to_numpy_rec_array(tree) if tree.parent.shape[0] > 0 else None
        for tree in condensed_trees
    ]
    linkage_trees = [tree if tree.shape[0] > 0 else None for tree in linkage_trees]

    # Aggregate the results
    (labels, probabilities, sub_labels, sub_probabilities, lens_values) = update_labels(
        cluster_probabilities,
        sub_labels,
        sub_probabilities,
        lens_values,
        points,
        data.shape[0],
    )

    # Propagate labels if requested
    if propagate_labels:
        labels, sub_labels = propagate_sub_cluster_labels(
            labels, sub_labels, core_graphs, points
        )

    # Reset for infinite data points
    if last_outlier > 0:
        (
            labels,
            probabilities,
            cluster_labels,
            cluster_probabilities,
            sub_labels,
            sub_probabilities,
            lens_values,
            points,
        ) = remap_results(
            labels,
            probabilities,
            cluster_labels,
            cluster_probabilities,
            sub_labels,
            sub_probabilities,
            lens_values,
            points,
            finite_index,
            clusterer._raw_data.shape[0],
        )

    return (
        labels,
        probabilities,
        cluster_labels,
        cluster_probabilities,
        sub_labels,
        sub_probabilities,
        core_graphs,
        condensed_trees,
        linkage_trees,
        spanning_trees,
        lens_values,
        points,
    )


class SubClusterDetector(ClusterMixin, BaseEstimator):
    """Performs a lens-value sub-cluster detection post-processing step
    on a HDBSCAN clusterer."""

    def __init__(
        self,
        *,
        lens_values=None,
        min_cluster_size=None,
        max_cluster_size=None,
        allow_single_cluster=False,
        cluster_selection_method="eom",
        cluster_selection_epsilon=0.0,
        cluster_selection_persistence=0.0,
        propagate_labels=False,
    ):
        self.lens_values = lens_values
        self.min_cluster_size = min_cluster_size
        self.max_cluster_size = max_cluster_size
        self.allow_single_cluster = allow_single_cluster
        self.cluster_selection_method = cluster_selection_method
        self.cluster_selection_epsilon = cluster_selection_epsilon
        self.cluster_selection_persistence = cluster_selection_persistence
        self.propagate_labels = propagate_labels

    def fit(
        self,
        clusterer,
        labels=None,
        probabilities=None,
        sample_weight=None,
        lens_callback=None,
    ):
        """labels and probabilities override the clusterer's values."""
        # get_params breaks with inherited classes!
        (
            self.labels_,
            self.probabilities_,
            self.cluster_labels_,
            self.cluster_probabilities_,
            self.sub_cluster_labels_,
            self.sub_cluster_probabilities_,
            self._approximation_graphs,
            self._condensed_trees,
            self._linkage_trees,
            self._spanning_trees,
            self.lens_values_,
            self.cluster_points_,
        ) = find_sub_clusters(
            clusterer,
            labels,
            probabilities,
            sample_weight,
            self.lens_values if lens_callback is None else lens_callback,
            min_cluster_size=self.min_cluster_size,
            max_cluster_size=self.max_cluster_size,
            allow_single_cluster=self.allow_single_cluster,
            cluster_selection_method=self.cluster_selection_method,
            cluster_selection_epsilon=self.cluster_selection_epsilon,
            cluster_selection_persistence=self.cluster_selection_persistence,
            propagate_labels=self.propagate_labels,
        )
        # also store the core distances and raw data for the member functions
        self._raw_data = clusterer._raw_data
        self._core_distances = clusterer._core_distances
        return self

    def propagated_labels(self):
        """Propagate sub-cluster labels to noise points."""
        check_is_fitted(
            self,
            "labels_",
            msg="You first need to fit the SubClusterDetector model before propagating the labels.",
        )
        return propagate_sub_cluster_labels(
            self.labels_.copy(),
            self.sub_cluster_labels_.copy(),
            self._approximation_graphs,
            self.cluster_points_,
        )

    @property
    def approximation_graph_(self):
        """See :class:`~hdbscan.plots.ApproximationGraph` for documentation."""
        return self._make_approximation_graph()

    def _make_approximation_graph(self, lens_name=None, sub_cluster_name=None):
        from hdbscan.plots import ApproximationGraph

        check_is_fitted(
            self,
            "_approximation_graphs",
            msg="You first need to fit the BranchDetector model before accessing the approximation graphs",
        )

        edge_lists = []
        for graph, points in zip(self._approximation_graphs, self.cluster_points_):
            edges = core_graph_to_edge_list(graph)
            edges[:, 0] = points[edges[:, 0].astype(np.int64)]
            edges[:, 1] = points[edges[:, 1].astype(np.int64)]
            edge_lists.append(edges)

        return ApproximationGraph(
            edge_lists,
            self.labels_,
            self.probabilities_,
            self.lens_values_,
            self.cluster_labels_,
            self.cluster_probabilities_,
            self.sub_cluster_labels_,
            self.sub_cluster_probabilities_,
            lens_name=lens_name,
            sub_cluster_name=sub_cluster_name,
            raw_data=self._raw_data,
        )

    @property
    def condensed_trees_(self):
        """See :class:`~hdbscan.plots.CondensedTree` for documentation."""
        from hdbscan.plots import CondensedTree

        check_is_fitted(
            self,
            "_condensed_trees",
            msg="You first need to fit the BranchDetector model before accessing the condensed trees",
        )
        return [
            (
                CondensedTree(
                    tree,
                    self.sub_cluster_labels_[points],
                )
                if tree is not None
                else None
            )
            for tree, points in zip(
                self._condensed_trees,
                self.cluster_points_,
            )
        ]

    @property
    def linkage_trees_(self):
        """See :class:`~hdbscan.plots.SingleLinkageTree` for documentation."""
        from hdbscan.plots import SingleLinkageTree

        check_is_fitted(
            self,
            "_linkage_trees",
            msg="You first need to fit the BranchDetector model before accessing the linkage trees",
        )
        return [
            SingleLinkageTree(tree) if tree is not None else None
            for tree in self._linkage_trees
        ]

    @property
    def spanning_trees_(self):
        """See :class:`~hdbscan.plots.MinimumSpanningTree` for documentation."""
        from hdbscan.plots import MinimumSpanningTree

        check_is_fitted(
            self,
            "_spanning_trees",
            msg="You first need to fit the BranchDetector model before accessing the linkage trees",
        )
        return [
            MinimumSpanningTree(tree, self._raw_data) for tree in self._spanning_trees
        ]
